@@ -17,9 +17,9 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with deluge.    If not, write to:
-# 	The Free Software Foundation, Inc.,
-# 	51 Franklin Street, Fifth Floor
-# 	Boston, MA  02110-1301, USA.
+#     The Free Software Foundation, Inc.,
+#     51 Franklin Street, Fifth Floor
+#     Boston, MA  02110-1301, USA.
 #
 #    In addition, as a special exception, the copyright holders give
 #    permission to link the code of portions of this program with the OpenSSL
@@ -42,11 +42,14 @@ import shutil
 import threading
 import pkg_resources
 import warnings
+import tempfile
+from urlparse import urljoin
 
 
 from twisted.internet import reactor, defer
 from twisted.internet.task import LoopingCall
 import twisted.web.client
+import twisted.web.error
 
 from deluge.httpdownloader import download_file
 from deluge.log import LOG as log
@@ -76,17 +79,15 @@ class Core(component.Component):
         # Start the libtorrent session
         log.info("Starting libtorrent %s session..", lt.version)
 
-        # We must depend on libtorrent >= 0.14.9 due to over-downloading bug
-        if lt.version < "0.14.9":
-            log.error("This version of Deluge requires libtorrent >= 0.14.9.")
-            sys.exit(1)
-            
         # Create the client fingerprint
         version = [int(value.split("-")[0]) for value in deluge.common.get_version().split(".")]
         while len(version) < 4:
             version.append(0)
 
-        self.session = lt.session(lt.fingerprint("DE", *version), flags=0)
+        # Note: All libtorrent python bindings to set plugins/extensions need to be disabled
+        # due to  GIL issue. https://code.google.com/p/libtorrent/issues/detail?id=369
+        # Setting session flags to 1 enables all libtorrent default plugins
+        self.session = lt.session(lt.fingerprint("DE", *version), flags=1)
 
         # Load the session state if available
         self.__load_session_state()
@@ -94,15 +95,24 @@ class Core(component.Component):
         # Set the user agent
         self.settings = lt.session_settings()
         self.settings.user_agent = "Deluge %s" % deluge.common.get_version()
+        # Increase the alert queue size so that alerts don't get lost
+        self.settings.alert_queue_size = 10000
 
         # Set session settings
         self.settings.send_redundant_have = True
+        if deluge.common.windows_check():
+            self.settings.disk_io_write_mode = \
+                lt.io_buffer_mode_t.disable_os_cache
+            self.settings.disk_io_read_mode = \
+                lt.io_buffer_mode_t.disable_os_cache
         self.session.set_settings(self.settings)
 
         # Load metadata extension
-        self.session.add_extension(lt.create_metadata_plugin)
-        self.session.add_extension(lt.create_ut_metadata_plugin)
-        self.session.add_extension(lt.create_smart_ban_plugin)
+        # Note: All libtorrent python bindings to set plugins/extensions need to be disabled
+        # due to  GIL issue. https://code.google.com/p/libtorrent/issues/detail?id=369
+        # self.session.add_extension(lt.create_metadata_plugin)
+        # self.session.add_extension(lt.create_ut_metadata_plugin)
+        # self.session.add_extension(lt.create_smart_ban_plugin)
 
         # Create the components
         self.eventmanager = EventManager()
@@ -124,8 +134,11 @@ class Core(component.Component):
         # store the one in the config so we can restore it on shutdown
         self.__old_interface = None
         if listen_interface:
-            self.__old_interface = self.config["listen_interface"]
-            self.config["listen_interface"] = listen_interface
+            if deluge.common.is_ip(listen_interface):
+                self.__old_interface = self.config["listen_interface"]
+                self.config["listen_interface"] = listen_interface
+            else:
+                log.error("Invalid listen interface (must be IP Address): %s", listen_interface)
 
     def start(self):
         """Starts the core"""
@@ -133,9 +146,12 @@ class Core(component.Component):
         self.__new_release = None
 
     def stop(self):
+        log.debug("Core stopping...")
+
         # Save the DHT state if necessary
         if self.config["dht"]:
             self.save_dht_state()
+
         # Save the libtorrent session state
         self.__save_session_state()
 
@@ -152,8 +168,9 @@ class Core(component.Component):
     def __save_session_state(self):
         """Saves the libtorrent session state"""
         try:
-            open(deluge.configmanager.get_config_dir("session.state"), "wb").write(
-                lt.bencode(self.session.state()))
+            lt_data = open(deluge.configmanager.get_config_dir("session.state"), "wb")
+            lt_data.write(lt.bencode(self.session.save_state()))
+            lt_data.close()
         except Exception, e:
             log.warning("Failed to save lt state: %s", e)
 
@@ -241,20 +258,35 @@ class Core(component.Component):
         :returns: a Deferred which returns the torrent_id as a str or None
         """
         log.info("Attempting to add url %s", url)
-        def on_get_file(filename):
+        def on_download_success(filename):
             # We got the file, so add it to the session
-            data = open(filename, "rb").read()
+            f = open(filename, "rb")
+            data = f.read()
+            f.close()
+            try:
+                os.remove(filename)
+            except Exception, e:
+                log.warning("Couldn't remove temp file: %s", e)
             return self.add_torrent_file(filename, base64.encodestring(data), options)
 
-        def on_get_file_error(failure):
-            # Log the error and pass the failure onto the client
-            log.error("Error occured downloading torrent from %s", url)
-            log.error("Reason: %s", failure.getErrorMessage())
-            return failure
+        def on_download_fail(failure):
+            if failure.check(twisted.web.error.PageRedirect):
+                new_url = urljoin(url, failure.getErrorMessage().split(" to ")[1])
+                result = download_file(new_url, tempfile.mkstemp()[1], headers=headers, force_filename=True)
+                result.addCallbacks(on_download_success, on_download_fail)
+            elif failure.check(twisted.web.client.PartialDownloadError):
+                result = download_file(url, tempfile.mkstemp()[1], headers=headers, force_filename=True,
+                        allow_compression=False)
+                result.addCallbacks(on_download_success, on_download_fail)
+            else:
+                # Log the error and pass the failure onto the client
+                log.error("Error occurred downloading torrent from %s", url)
+                log.error("Reason: %s", failure.getErrorMessage())
+                result = failure
+            return result
 
-        d = download_file(url, url.split("/")[-1], headers=headers)
-        d.addCallback(on_get_file)
-        d.addErrback(on_get_file_error)
+        d = download_file(url, tempfile.mkstemp()[1], headers=headers, force_filename=True)
+        d.addCallbacks(on_download_success, on_download_fail)
         return d
 
     @export
@@ -293,28 +325,6 @@ class Core(component.Component):
         log.debug("Removing torrent %s from the core.", torrent_id)
         return self.torrentmanager.remove(torrent_id, remove_data)
 
-    @export
-    def get_stats(self):
-        """
-        Deprecated: please use get_session_status()
-
-        """
-        warnings.warn("Use get_session_status() instead of get_stats()", DeprecationWarning)
-        stats = self.get_session_status(["payload_download_rate", "payload_upload_rate",
-            "dht_nodes", "has_incoming_connections", "download_rate", "upload_rate"])
-
-        stats.update({
-            #dynamic stats:
-            "num_connections":self.session.num_connections(),
-            "free_space":deluge.common.free_space(self.config["download_location"]),
-            #max config values:
-            "max_download":self.config["max_download_speed"],
-            "max_upload":self.config["max_upload_speed"],
-            "max_num_connections":self.config["max_connections_global"],
-        })
-
-        return stats
-
     @export
     def get_session_status(self, keys):
         """
@@ -412,9 +422,13 @@ class Core(component.Component):
             self.torrentmanager[torrent_id].resume()
 
     @export
-    def get_torrent_status(self, torrent_id, keys):
+    def get_torrent_status(self, torrent_id, keys, diff=False):
         # Build the status dictionary
-        status = self.torrentmanager[torrent_id].get_status(keys)
+        try:
+            status = self.torrentmanager[torrent_id].get_status(keys, diff)
+        except KeyError:
+            # Torrent was probaly removed meanwhile
+            return {}
 
         # Get the leftover fields and ask the plugin manager to fill them
         leftover_fields = list(set(keys) - set(status.keys()))
@@ -423,7 +437,7 @@ class Core(component.Component):
         return status
 
     @export
-    def get_torrents_status(self, filter_dict, keys):
+    def get_torrents_status(self, filter_dict, keys, diff=False):
         """
         returns all torrents , optionally filtered by filter_dict.
         """
@@ -432,7 +446,7 @@ class Core(component.Component):
 
         # Get the torrent status for each torrent_id
         for torrent_id in torrent_ids:
-            status_dict[torrent_id] = self.get_torrent_status(torrent_id, keys)
+            status_dict[torrent_id] = self.get_torrent_status(torrent_id, keys, diff)
 
         return status_dict
 
@@ -473,7 +487,7 @@ class Core(component.Component):
             try:
                 config[key] = self.config[key]
             except KeyError:
-                pass
+                config[key] = None
         return config
 
     @export
@@ -495,24 +509,6 @@ class Core(component.Component):
         """Returns the current number of connections"""
         return self.session.num_connections()
 
-    @export
-    def get_dht_nodes(self):
-        """Returns the number of dht nodes"""
-        warnings.warn("Use get_session_status().", DeprecationWarning)
-        return self.session.status().dht_nodes
-
-    @export
-    def get_download_rate(self):
-        """Returns the payload download rate"""
-        warnings.warn("Use get_session_status().", DeprecationWarning)
-        return self.session.status().payload_download_rate
-
-    @export
-    def get_upload_rate(self):
-        """Returns the payload upload rate"""
-        warnings.warn("Use get_session_status().", DeprecationWarning)
-        return self.session.status().payload_upload_rate
-
     @export
     def get_available_plugins(self):
         """Returns a list of plugins available in the core"""
@@ -610,12 +606,6 @@ class Core(component.Component):
         """Sets the path for the torrent to be moved when completed"""
         return self.torrentmanager[torrent_id].set_move_completed_path(value)
 
-    @export
-    def get_health(self):
-        """Returns True if we have established incoming connections"""
-        warnings.warn("Use get_session_status().", DeprecationWarning)
-        return self.session.status().has_incoming_connections
-
     @export
     def get_path_size(self, path):
         """Returns the size of the file or folder 'path' and -1 if the path is
@@ -627,7 +617,7 @@ class Core(component.Component):
                         webseeds, private, created_by, trackers, add_to_session):
 
         log.debug("creating torrent..")
-        threading.Thread(target=_create_torrent_thread,
+        threading.Thread(target=self._create_torrent_thread,
             args=(
                 path,
                 tracker,
@@ -730,7 +720,8 @@ class Core(component.Component):
     @export
     def queue_top(self, torrent_ids):
         log.debug("Attempting to queue %s to top", torrent_ids)
-        for torrent_id in torrent_ids:
+        # torrent_ids must be sorted in reverse before moving to preserve order
+        for torrent_id in sorted(torrent_ids, key=self.torrentmanager.get_queue_position, reverse=True):
             try:
                 # If the queue method returns True, then we should emit a signal
                 if self.torrentmanager.queue_top(torrent_id):
@@ -741,35 +732,48 @@ class Core(component.Component):
     @export
     def queue_up(self, torrent_ids):
         log.debug("Attempting to queue %s to up", torrent_ids)
+        torrents = ((self.torrentmanager.get_queue_position(torrent_id), torrent_id) for torrent_id in torrent_ids)
+        torrent_moved = True
+        prev_queue_position = None
         #torrent_ids must be sorted before moving.
-        torrent_ids = list(torrent_ids)
-        torrent_ids.sort(key = lambda id: self.torrentmanager.torrents[id].get_queue_position())
-        for torrent_id in torrent_ids:
-            try:
-                # If the queue method returns True, then we should emit a signal
-                if self.torrentmanager.queue_up(torrent_id):
-                    component.get("EventManager").emit(TorrentQueueChangedEvent())
-            except KeyError:
-                log.warning("torrent_id: %s does not exist in the queue", torrent_id)
+        for queue_position, torrent_id in sorted(torrents):
+            # Move the torrent if and only if there is space (by not moving it we preserve the order)
+            if torrent_moved or queue_position - prev_queue_position > 1:
+                try:
+                    torrent_moved = self.torrentmanager.queue_up(torrent_id)
+                except KeyError:
+                    log.warning("torrent_id: %s does not exist in the queue", torrent_id)
+            # If the torrent moved, then we should emit a signal
+            if torrent_moved:
+                component.get("EventManager").emit(TorrentQueueChangedEvent())
+            else:
+                prev_queue_position = queue_position
 
     @export
     def queue_down(self, torrent_ids):
         log.debug("Attempting to queue %s to down", torrent_ids)
+        torrents = ((self.torrentmanager.get_queue_position(torrent_id), torrent_id) for torrent_id in torrent_ids)
+        torrent_moved = True
+        prev_queue_position = None
         #torrent_ids must be sorted before moving.
-        torrent_ids = list(torrent_ids)
-        torrent_ids.sort(key = lambda id: -self.torrentmanager.torrents[id].get_queue_position())
-        for torrent_id in torrent_ids:
-            try:
-                # If the queue method returns True, then we should emit a signal
-                if self.torrentmanager.queue_down(torrent_id):
-                    component.get("EventManager").emit(TorrentQueueChangedEvent())
-            except KeyError:
-                log.warning("torrent_id: %s does not exist in the queue", torrent_id)
+        for queue_position, torrent_id in sorted(torrents, reverse=True):
+            # Move the torrent if and only if there is space (by not moving it we preserve the order)
+            if torrent_moved or prev_queue_position - queue_position > 1:
+                try:
+                    torrent_moved = self.torrentmanager.queue_down(torrent_id)
+                except KeyError:
+                    log.warning("torrent_id: %s does not exist in the queue", torrent_id)
+            # If the torrent moved, then we should emit a signal
+            if torrent_moved:
+                component.get("EventManager").emit(TorrentQueueChangedEvent())
+            else:
+                prev_queue_position = queue_position
 
     @export
     def queue_bottom(self, torrent_ids):
         log.debug("Attempting to queue %s to bottom", torrent_ids)
-        for torrent_id in torrent_ids:
+        # torrent_ids must be sorted before moving to preserve order
+        for torrent_id in sorted(torrent_ids, key=self.torrentmanager.get_queue_position):
             try:
                 # If the queue method returns True, then we should emit a signal
                 if self.torrentmanager.queue_bottom(torrent_id):
@@ -792,21 +796,27 @@ class Core(component.Component):
         """
         from twisted.web.client import getPage
 
-        d = getPage("http://deluge-torrent.org/test_port.php?port=%s" % self.get_listen_port())
+        d = getPage("http://deluge-torrent.org/test_port.php?port=%s" %
+                    self.get_listen_port(), timeout=30)
 
         def on_get_page(result):
             return bool(int(result))
 
+        def logError(failure):
+            log.warning("Error testing listen port: %s", failure)
+
         d.addCallback(on_get_page)
+        d.addErrback(logError)
 
         return d
 
     @export
-    def get_free_space(self, path):
+    def get_free_space(self, path=None):
         """
         Returns the number of free bytes at path
 
-        :param path: the path to check free space at
+        :param path: the path to check free space at, if None, use the default
+        download location
         :type path: string
 
         :returns: the number of free bytes at path
@@ -815,7 +825,12 @@ class Core(component.Component):
         :raises InvalidPathError: if the path is invalid
 
         """
-        return deluge.common.free_space(path)
+        if not path:
+            path = self.config["download_location"]
+        try:
+            return deluge.common.free_space(path)
+        except InvalidPathError:
+            return 0
 
     @export
     def get_libtorrent_version(self):
