@@ -17,9 +17,9 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with deluge.    If not, write to:
-# 	The Free Software Foundation, Inc.,
-# 	51 Franklin Street, Fifth Floor
-# 	Boston, MA  02110-1301, USA.
+#   The Free Software Foundation, Inc.,
+#   51 Franklin Street, Fifth Floor
+#   Boston, MA  02110-1301, USA.
 #
 #    In addition, as a special exception, the copyright holders give
 #    permission to link the code of portions of this program with the OpenSSL
@@ -138,6 +138,7 @@ class TorrentManager(component.Component):
 
         # Create the torrents dict { torrent_id: Torrent }
         self.torrents = {}
+        self.queued_torrents = set()
 
         # This is a list of torrent_id when we shutdown the torrentmanager.
         # We use this list to determine if all active torrents have been paused
@@ -149,7 +150,7 @@ class TorrentManager(component.Component):
 
         # Keeps track of resume data that needs to be saved to disk
         self.resume_data = {}
-        
+
         # Workaround to determine if TorrentAddedEvent is from state file
         self.session_started = False
 
@@ -375,14 +376,20 @@ class TorrentManager(component.Component):
                 resume_data = self.legacy_get_resume_data_from_file(state.torrent_id)
                 self.legacy_delete_resume_data(state.torrent_id)
 
-            add_torrent_params["resume_data"] = resume_data
+            if resume_data:
+                add_torrent_params["resume_data"] = resume_data
         else:
             # We have a torrent_info object or magnet uri so we're not loading from state.
             if torrent_info:
                 add_torrent_id = str(torrent_info.info_hash())
+                # If this torrent id is already in the session, merge any additional trackers.
                 if add_torrent_id in self.get_torrent_list():
-                    # Torrent already exists just append any extra trackers.
-                    log.debug("Torrent (%s) exists, checking for trackers to add...", add_torrent_id)
+                    log.info("Merging trackers for torrent (%s) already in session...", add_torrent_id)
+                    # Don't merge trackers if either torrent has private flag set
+                    if self[add_torrent_id].get_status(["private"])["private"]:
+                        log.info("Merging trackers abandoned: Torrent has private flag set.")
+                        return
+
                     add_torrent_trackers = []
                     for value in torrent_info.trackers():
                         tracker = {}
@@ -392,17 +399,18 @@ class TorrentManager(component.Component):
 
                     torrent_trackers = {}
                     tracker_list = []
-                    for tracker in  self[add_torrent_id].get_status(["trackers"])["trackers"]:
+                    for tracker in self[add_torrent_id].get_status(["trackers"])["trackers"]:
                         torrent_trackers[(tracker["url"])] = tracker
                         tracker_list.append(tracker)
 
-                    added_tracker = False
+                    added_tracker = 0
                     for tracker in add_torrent_trackers:
                         if tracker['url'] not in torrent_trackers:
                             tracker_list.append(tracker)
-                            added_tracker = True
+                            added_tracker += 1
 
                     if added_tracker:
+                        log.info("%s tracker(s) merged into torrent.", added_tracker)
                         self[add_torrent_id].set_trackers(tracker_list)
                     return
 
@@ -417,12 +425,19 @@ class TorrentManager(component.Component):
             # Check for renamed files and if so, rename them in the torrent_info
             # before adding to the session.
             if options["mapped_files"]:
-                for index, name in options["mapped_files"].items():
-                    log.debug("renaming file index %s to %s", index, name)
-                    torrent_info.rename_file(index, utf8_encoded(name))
+                for index, fname in options["mapped_files"].items():
+                    try:
+                        fname = unicode(fname, "utf-8")
+                    except TypeError:
+                        pass
+                    fname = deluge.core.torrent.sanitize_filepath(fname)
+                    log.debug("renaming file index %s to %s", index, fname)
+                    try:
+                        torrent_info.rename_file(index, fname)
+                    except TypeError:
+                        torrent_info.rename_file(index, fname.encode("utf-8"))
 
             add_torrent_params["ti"] = torrent_info
-            add_torrent_params["resume_data"] = ""
 
         #log.info("Adding torrent: %s", filename)
         log.debug("options: %s", options)
@@ -447,7 +462,7 @@ class TorrentManager(component.Component):
         handle = None
         try:
             if magnet:
-                handle = lt.add_magnet_uri(self.session, magnet, add_torrent_params)
+                handle = lt.add_magnet_uri(self.session, utf8_encoded(magnet), add_torrent_params)
             else:
                 handle = self.session.add_torrent(add_torrent_params)
         except RuntimeError, e:
@@ -475,6 +490,9 @@ class TorrentManager(component.Component):
         if not options["add_paused"]:
             torrent.resume()
 
+        # Add to queued torrents set
+        self.queued_torrents.add(torrent.torrent_id)
+
         # Write the .torrent file to the state directory
         if filedump:
             try:
@@ -541,12 +559,11 @@ class TorrentManager(component.Component):
         :raises InvalidTorrentError: if the torrent_id is not in the session
 
         """
-
-        if torrent_id not in self.torrents:
+        try:
+            torrent_name = self.torrents[torrent_id].get_status(["name"])["name"]
+        except KeyError:
             raise InvalidTorrentError("torrent_id not in session")
 
-        torrent_name = self.torrents[torrent_id].get_status(["name"])["name"]
-
         # Emit the signal to the clients
         component.get("EventManager").emit(PreTorrentRemovedEvent(torrent_id))
 
@@ -583,10 +600,17 @@ class TorrentManager(component.Component):
         # Stop the looping call
         self.torrents[torrent_id].prev_status_cleanup_loop.stop()
 
+        # Remove from set if it wasn't finished
+        if not self.torrents[torrent_id].is_finished:
+            try:
+                self.queued_torrents.remove(torrent_id)
+            except KeyError:
+                log.debug("%s isn't in queued torrents set?", torrent_id)
+
         # Remove the torrent from deluge's session
         try:
             del self.torrents[torrent_id]
-        except KeyError, ValueError:
+        except (KeyError, ValueError):
             return False
 
         # Save the session state
@@ -607,7 +631,7 @@ class TorrentManager(component.Component):
                 os.path.join(get_config_dir(), "state", "torrents.state"), "rb")
             state = cPickle.load(state_file)
             state_file.close()
-        except (EOFError, IOError, Exception), e:
+        except (EOFError, IOError, Exception, cPickle.UnpicklingError), e:
             log.warning("Unable to load state file: %s", e)
 
         # Try to use an old state
@@ -622,7 +646,7 @@ class TorrentManager(component.Component):
 
         # Reorder the state.torrents list to add torrents in the correct queue
         # order.
-        state.torrents.sort(key=operator.attrgetter("queue"))
+        state.torrents.sort(key=operator.attrgetter("queue"), reverse=self.config["queue_new_to_top"])
 
         resume_data = self.load_resume_data_file()
 
@@ -682,8 +706,8 @@ class TorrentManager(component.Component):
             state_file.flush()
             os.fsync(state_file.fileno())
             state_file.close()
-        except IOError:
-            log.warning("Unable to save state file.")
+        except IOError, e:
+            log.warning("Unable to save state file: %s", e)
             return True
 
         # We have to move the 'torrents.state.new' file to 'torrents.state'
@@ -770,10 +794,10 @@ class TorrentManager(component.Component):
         Cleans up after libtorrent folder renames.
 
         """
-        if torrent_id not in self.torrents:
-            raise InvalidTorrentError("torrent_id is not in session")
-
-        info = self.torrents[torrent_id].get_status(['save_path'])
+        try:
+            info = self.torrents[torrent_id].get_status(['save_path'])
+        except KeyError:
+            raise InvalidTorrentError("torrent_id not in session")
         # Regex removes leading slashes that causes join function to ignore save_path
         folder_full_path = os.path.join(info['save_path'], re.sub("^/*", "", folder))
         folder_full_path = os.path.normpath(folder_full_path)
@@ -789,13 +813,18 @@ class TorrentManager(component.Component):
                             os.removedirs(os.path.join(root, name))
                             log.debug("Removed Empty Folder %s", os.path.join(root, name))
                         except OSError, (errno, strerror):
-                            if errno == 39:
+                            from errno import ENOTEMPTY
+                            if errno == ENOTEMPTY:
                                 # Error raised if folder is not empty
                                 log.debug("%s", strerror)
 
         except OSError, (errno, strerror):
             log.debug("Cannot Remove Folder: %s (ErrNo %s)", strerror, errno)
 
+    def get_queue_position(self, torrent_id):
+        """Get queue position of torrent"""
+        return self.torrents[torrent_id].get_queue_position()
+
     def queue_top(self, torrent_id):
         """Queue torrent to top"""
         if self.torrents[torrent_id].get_queue_position() == 0:
@@ -814,7 +843,7 @@ class TorrentManager(component.Component):
 
     def queue_down(self, torrent_id):
         """Queue torrent down one position"""
-        if self.torrents[torrent_id].get_queue_position() == (len(self.torrents) - 1):
+        if self.torrents[torrent_id].get_queue_position() == (len(self.queued_torrents) - 1):
             return False
 
         self.torrents[torrent_id].handle.queue_position_down()
@@ -822,7 +851,7 @@ class TorrentManager(component.Component):
 
     def queue_bottom(self, torrent_id):
         """Queue torrent to bottom"""
-        if self.torrents[torrent_id].get_queue_position() == (len(self.torrents) - 1):
+        if self.torrents[torrent_id].get_queue_position() == (len(self.queued_torrents) - 1):
             return False
 
         self.torrents[torrent_id].handle.queue_position_bottom()
@@ -855,9 +884,9 @@ class TorrentManager(component.Component):
         log.debug("on_alert_torrent_finished")
         try:
             torrent = self.torrents[str(alert.handle.info_hash())]
+            torrent_id = str(alert.handle.info_hash())
         except:
             return
-        torrent_id = str(alert.handle.info_hash())
         log.debug("%s is finished..", torrent_id)
 
         # Get the total_download and if it's 0, do not move.. It's likely
@@ -873,10 +902,19 @@ class TorrentManager(component.Component):
                 if torrent.options["download_location"] != move_path:
                     torrent.move_storage(move_path)
 
+        torrent.update_state()
+        if not torrent.is_finished and total_download:
             torrent.is_finished = True
             component.get("EventManager").emit(TorrentFinishedEvent(torrent_id))
+        else:
+            torrent.is_finished = True
 
-        torrent.update_state()
+        # Torrent is no longer part of the queue
+        try:
+            self.queued_torrents.remove(torrent_id)
+        except KeyError:
+            # Sometimes libtorrent fires a TorrentFinishedEvent twice
+            log.debug("%s isn't in queued torrents set?", torrent_id)
 
         # Only save resume data if it was actually downloaded something. Helps
         # on startup with big queues with lots of seeding torrents. Libtorrent
@@ -890,9 +928,9 @@ class TorrentManager(component.Component):
         log.debug("on_alert_torrent_paused")
         try:
             torrent = self.torrents[str(alert.handle.info_hash())]
+            torrent_id = str(alert.handle.info_hash())
         except:
             return
-        torrent_id = str(alert.handle.info_hash())
         # Set the torrent state
         old_state = torrent.state
         torrent.update_state()
@@ -984,10 +1022,9 @@ class TorrentManager(component.Component):
         log.debug("on_alert_torrent_resumed")
         try:
             torrent = self.torrents[str(alert.handle.info_hash())]
+            torrent_id = str(alert.handle.info_hash())
         except:
             return
-        torrent_id = str(alert.handle.info_hash())
-        torrent.is_finished = torrent.handle.is_seed()
         old_state = torrent.state
         torrent.update_state()
         if torrent.state != old_state:
@@ -1005,16 +1042,20 @@ class TorrentManager(component.Component):
 
         old_state = torrent.state
         torrent.update_state()
+
+        # Torrent may need to download data after checking.
+        if torrent.state in ('Checking', 'Checking Resume Data', 'Downloading'):
+            torrent.is_finished = False
+            self.queued_torrents.add(torrent_id)
+
         # Only emit a state changed event if the state has actually changed
         if torrent.state != old_state:
             component.get("EventManager").emit(TorrentStateChangedEvent(torrent_id, torrent.state))
 
     def on_alert_save_resume_data(self, alert):
         log.debug("on_alert_save_resume_data")
-
-        torrent_id = str(alert.handle.info_hash())
-
         try:
+            torrent_id = str(alert.handle.info_hash())
             torrent = self.torrents[torrent_id]
         except:
             return
@@ -1045,9 +1086,9 @@ class TorrentManager(component.Component):
         log.debug("index: %s name: %s", alert.index, alert.name.decode("utf8"))
         try:
             torrent = self.torrents[str(alert.handle.info_hash())]
+            torrent_id = str(alert.handle.info_hash())
         except:
             return
-        torrent_id = str(alert.handle.info_hash())
 
         # We need to see if this file index is in a waiting_on_folder list
         folder_rename = False
@@ -1077,7 +1118,7 @@ class TorrentManager(component.Component):
             torrent = self.torrents[str(alert.handle.info_hash())]
         except:
             return
-        torrent.write_torrentfile()
+        torrent.on_metadata_received()
 
     def on_alert_file_error(self, alert):
         log.debug("on_alert_file_error: %s", alert.message())
@@ -1089,6 +1130,9 @@ class TorrentManager(component.Component):
 
     def on_alert_file_completed(self, alert):
         log.debug("file_completed_alert: %s", alert.message())
-        torrent_id = str(alert.handle.info_hash())
+        try:
+            torrent_id = str(alert.handle.info_hash())
+        except:
+            return
         component.get("EventManager").emit(
             TorrentFileCompletedEvent(torrent_id, alert.index))
