@@ -18,9 +18,9 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with deluge.    If not, write to:
-# 	The Free Software Foundation, Inc.,
-# 	51 Franklin Street, Fifth Floor
-# 	Boston, MA  02110-1301, USA.
+#     The Free Software Foundation, Inc.,
+#     51 Franklin Street, Fifth Floor
+#     Boston, MA  02110-1301, USA.
 #
 #    In addition, as a special exception, the copyright holders give
 #    permission to link the code of portions of this program with the OpenSSL
@@ -48,6 +48,12 @@ import deluge.configmanager
 import deluge.common
 from deluge.log import LOG as log
 
+from gobject import signal_new, SIGNAL_RUN_LAST, TYPE_NONE
+from gtk import gdk
+signal_new('button-press-event', gtk.TreeViewColumn,
+           SIGNAL_RUN_LAST, TYPE_NONE, (gdk.Event,))
+
+
 # Cell data functions to pass to add_func_column()
 def cell_data_speed(column, cell, model, row, data):
     """Display value as a speed, eg. 2 KiB/s"""
@@ -96,6 +102,15 @@ def cell_data_date(column, cell, model, row, data):
     """Display value as date, eg 05/05/08"""
     cell.set_property('text', deluge.common.fdate(model.get_value(row, data)))
 
+def cell_data_speed_limit(column, cell, model, row, data):
+    """Display value as a speed, eg. 2 KiB/s"""
+    speed = model.get_value(row, data)
+    speed_str = ""
+    if speed > 0:
+        speed_str = deluge.common.fspeed(speed * 1024)
+
+    cell.set_property('text', speed_str)
+
 class ListViewColumnState:
     """Used for saving/loading column state"""
     def __init__(self, name, position, width, visible, sort, sort_order):
@@ -132,6 +147,32 @@ class ListView:
             self.sort_func = None
             self.sort_id = None
 
+    class TreeviewColumn(gtk.TreeViewColumn):
+        """
+            TreeViewColumn does not signal right-click events, and we need them
+            This subclass is equivalent to TreeViewColumn, but it signals these events
+
+            Most of the code of this class comes from Quod Libet (http://www.sacredchao.net/quodlibet)
+        """
+
+        def __init__(self, title=None, cell_renderer=None, ** args):
+            """ Constructor, see gtk.TreeViewColumn """
+            gtk.TreeViewColumn.__init__(self, title, cell_renderer, ** args)
+            label = gtk.Label(title)
+            self.set_widget(label)
+            label.show()
+            label.__realize = label.connect('realize', self.onRealize)
+
+        def onRealize(self, widget):
+            widget.disconnect(widget.__realize)
+            del widget.__realize
+            button = widget.get_ancestor(gtk.Button)
+            if button is not None:
+                button.connect('button-press-event', self.onButtonPressed)
+
+        def onButtonPressed(self, widget, event):
+            self.emit('button-press-event', event)
+
     def __init__(self, treeview_widget=None, state_file=None):
         log.debug("ListView initialized..")
 
@@ -141,6 +182,9 @@ class ListView:
         else:
             self.treeview = gtk.TreeView()
 
+        self.treeview.set_enable_search(True)
+        self.treeview.set_search_equal_func(self.on_keypress_search_by_name)
+
         if state_file:
             self.load_state(state_file)
 
@@ -148,7 +192,7 @@ class ListView:
         self.model_filter = None
 
         self.treeview.set_rules_hint(True)
-        self.treeview.set_reorderable(True)
+        self.treeview.set_reorderable(False)
         self.treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
 
         # Dictionary of 'header' or 'name' to ListViewColumn object
@@ -189,13 +233,25 @@ class ListView:
         self.set_sort_functions()
         self.treeview.set_model(self.model_filter)
 
+    def stabilize_sort_func(self, sort_func):
+        def stabilized(model, iter1, iter2, data):
+            result = sort_func(model, iter1, iter2, data)
+            if result == 0:
+                # Compare by hash if main column values are equal
+                return cmp(model[iter1][1], model[iter2][1])
+            return result
+        return stabilized
+
+    def generic_sort_func(self, model, iter1, iter2, data):
+        return cmp(model[iter1][data], model[iter2][data])
+
     def set_sort_functions(self):
         for column in self.columns.values():
-            if column.sort_func:
-                self.model_filter.set_sort_func(
-                    column.sort_id,
-                    column.sort_func,
-                    column.sort_id)
+            sort_func = column.sort_func or self.generic_sort_func
+            self.model_filter.set_sort_func(
+                column.sort_id,
+                self.stabilize_sort_func(sort_func),
+                column.sort_id)
 
     def create_column_state(self, column, position=None):
         if not position:
@@ -219,6 +275,9 @@ class ListView:
         # A list of ListViewColumnStates
         state = []
 
+        # Workaround for all zero widths after removing column on shutdown
+        if not any(c.get_width() for c in self.treeview.get_columns()): return
+
         # Get the list of TreeViewColumns from the TreeView
         for counter, column in enumerate(self.treeview.get_columns()):
             # Append a new column state to the state list
@@ -248,7 +307,7 @@ class ListView:
             state_file = open(os.path.join(config_location, filename), "rb")
             state = cPickle.load(state_file)
             state_file.close()
-        except (EOFError, IOError), e:
+        except (EOFError, IOError, cPickle.UnpicklingError), e:
             log.warning("Unable to load state file: %s", e)
 
         # Keep the state in self.state so we can access it as we add new columns
@@ -289,9 +348,17 @@ class ListView:
         name = widget.get_child().get_text()
 
         # Set the column's visibility based on the widgets active state
-        self.columns[name].column.set_visible(widget.get_active())
+        try:
+            self.columns[name].column.set_visible(widget.get_active())
+        except KeyError:
+            self.columns[unicode(name)].column.set_visible(widget.get_active())
         return
 
+    def on_treeview_header_right_clicked(self, column, event):
+        if event.button == 3:
+            self.menu.popup(None, None, None, event.button, event.get_time())
+
+
     def register_checklist_menu(self, menu):
         """Register a checklist menu with the listview.  It will automatically
         attach any new checklist menu it makes to this menu.
@@ -300,7 +367,7 @@ class ListView:
 
     def create_checklist_menu(self):
         """Creates a menu used for toggling the display of columns."""
-        menu = gtk.Menu()
+        menu = self.menu = gtk.Menu()
         # Iterate through the column_index list to preserve order
         for name in self.column_index:
             column = self.columns[name]
@@ -322,7 +389,6 @@ class ListView:
         for _menu in self.checklist_menus:
             _menu.set_submenu(menu)
             _menu.show_all()
-
         return menu
 
     def create_new_liststore(self):
@@ -387,7 +453,7 @@ class ListView:
 
     def add_column(self, header, render, col_types, hidden, position,
             status_field, sortid, text=0, value=0, pixbuf=0, function=None,
-            column_type=None, sort_func=None):
+            column_type=None, sort_func=None, default=True):
         """Adds a column to the ListView"""
         # Add the column types to liststore_columns
         column_indices = []
@@ -415,7 +481,8 @@ class ListView:
         # Create a new list with the added column
         self.create_new_liststore()
 
-        column = gtk.TreeViewColumn(header)
+        column = self.TreeviewColumn(header)
+
         if column_type == "text":
             column.pack_start(render)
             column.add_attribute(render, "text",
@@ -460,12 +527,16 @@ class ListView:
         column.set_min_width(10)
         column.set_reorderable(True)
         column.set_visible(not hidden)
+        column.connect('button-press-event',
+                       self.on_treeview_header_right_clicked)
 
         # Check for loaded state and apply
+        column_in_state = False
         if self.state != None:
             for column_state in self.state:
                 if header == column_state.name:
                     # We found a loaded state
+                    column_in_state = True
                     if column_state.width > 0:
                         column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
                         column.set_fixed_width(column_state.width)
@@ -474,6 +545,12 @@ class ListView:
                         self.model_filter.set_sort_column_id(column_state.sort, column_state.sort_order)
                     column.set_visible(column_state.visible)
                     position = column_state.position
+                    break
+
+        # Set this column to not visible if its not in the state and
+        # its not supposed to be shown by default
+        if not column_in_state and not default and not hidden:
+            column.set_visible(False)
 
         if position is not None:
             self.treeview.insert_column(column, position)
@@ -488,75 +565,110 @@ class ListView:
 
         return True
 
-    def add_text_column(self, header, col_type=str, hidden=False,
-                                            position=None,
-                                            status_field=None,
-                                            sortid=0,
-                                            column_type="text",
-                                            sort_func=None):
+    def add_text_column(self, header, col_type=str, hidden=False, position=None,
+                        status_field=None, sortid=0, column_type="text",
+                        sort_func=None, default=True):
         """Add a text column to the listview.  Only the header name is required.
         """
         render = gtk.CellRendererText()
         self.add_column(header, render, col_type, hidden, position,
-                    status_field, sortid, column_type=column_type, sort_func=sort_func)
+                        status_field, sortid, column_type=column_type,
+                        sort_func=sort_func, default=default)
 
         return True
 
     def add_bool_column(self, header, col_type=bool, hidden=False,
-                                            position=None,
-                                            status_field=None,
-                                            sortid=0,
-                                            column_type="bool"):
-
+                        position=None, status_field=None, sortid=0,
+                        column_type="bool", default=True):
         """Add a bool column to the listview"""
         render = gtk.CellRendererToggle()
         self.add_column(header, render, col_type, hidden, position,
-                    status_field, sortid, column_type=column_type)
+                        status_field, sortid, column_type=column_type,
+                        default=default)
 
     def add_func_column(self, header, function, col_types, sortid=0,
-                                hidden=False, position=None, status_field=None,
-                                column_type="func", sort_func=None):
+                        hidden=False, position=None, status_field=None,
+                        column_type="func", sort_func=None, default=True):
         """Add a function column to the listview.  Need a header name, the
         function and the column types."""
 
         render = gtk.CellRendererText()
         self.add_column(header, render, col_types, hidden, position,
-                            status_field, sortid, column_type=column_type,
-                            function=function, sort_func=sort_func)
+                        status_field, sortid, column_type=column_type,
+                        function=function, sort_func=sort_func, default=default)
 
         return True
 
-    def add_progress_column(self, header, col_types=[float, str],
-                                            sortid=0,
-                                            hidden=False,
-                                            position=None,
-                                            status_field=None,
-                                            function=None,
-                                            column_type="progress"):
+    def add_progress_column(self, header, col_types=[float, str], sortid=0,
+                            hidden=False, position=None, status_field=None,
+                            function=None, column_type="progress",
+                            default=True):
         """Add a progress column to the listview."""
 
         render = gtk.CellRendererProgress()
         self.add_column(header, render, col_types, hidden, position,
-                            status_field, sortid, function=function,
-                            column_type=column_type,
-                            value=0, text=1)
+                        status_field, sortid, function=function,
+                        column_type=column_type, value=0, text=1,
+                        default=default)
 
         return True
 
-    def add_texticon_column(self, header, col_types=[str, str],
-                                            sortid=1,
-                                            hidden=False,
-                                            position=None,
-                                            status_field=None,
-                                            column_type="texticon",
-                                            function=None):
+    def add_texticon_column(self, header, col_types=[str, str], sortid=1,
+                            hidden=False, position=None, status_field=None,
+                            column_type="texticon", function=None,
+                            default=True):
         """Adds a texticon column to the listview."""
         render1 = gtk.CellRendererPixbuf()
         render2 = gtk.CellRendererText()
 
-        self.add_column(header, (render1, render2), col_types, hidden,
-                            position, status_field, sortid,
-                            column_type=column_type, function=function,
-                            pixbuf=0, text=1)
+        self.add_column(header, (render1, render2), col_types, hidden, position,
+                        status_field, sortid, column_type=column_type,
+                        function=function, pixbuf=0, text=1, default=default)
 
         return True
+
+    def on_keypress_search_by_name(self, model, columnn, key, iter):
+        TORRENT_NAME_COL = 5
+        return not model[iter][TORRENT_NAME_COL].lower().startswith(key.lower())
+
+    def restore_columns_order_from_state(self):
+        if self.state is None:
+            # No state file exists, so, no reordering can be done
+            return
+        columns = self.treeview.get_columns()
+        def find_column(header):
+            for column in columns:
+                if column.get_title() == header:
+                    return column
+
+        restored_columns = []
+        for col_state in self.state:
+            if col_state.name in restored_columns:
+                # Duplicate column in state!?!?!?
+                continue
+            elif not col_state.visible:
+                # Column is not visible, no need to reposition
+                continue
+
+            try:
+                column_at_position = columns[col_state.position]
+            except IndexError:
+                # Extra columns in loaded state, likely from plugins, so just skip them.
+                continue
+            if col_state.name == column_at_position.get_title():
+                # It's in the right position
+                continue
+            column = find_column(col_state.name)
+            if not column:
+                log.debug("Could not find column matching \"%s\" on state." %
+                          col_state.name)
+                # The cases where I've found that the column could not be found
+                # is when not using the english locale, ie, the default one, or
+                # when changing locales between runs.
+                # On the next load, all should be fine
+                continue
+            self.treeview.move_column_after(column, column_at_position)
+            # Get columns again to keep reordering since positions have changed
+            columns = self.treeview.get_columns()
+            restored_columns.append(col_state.name)
+        self.create_new_liststore()
